package flm.b4a.openslplayer;

import java.io.File;
import java.io.FileDescriptor;
import java.io.FileInputStream;
import java.io.IOException;

import android.content.res.AssetFileDescriptor;
import anywheresoftware.b4a.BA;

@BA.Author(value = "F. Leneuf-Magaud & J. Kytomaki")
@BA.ShortName("OpenSLSoundPlayer")
@BA.Version(value = 1.0F)
public class OpenSLSoundPlayer
{
	private static native void createEngine(int maxStreams, int channels);
	private static native int load(FileDescriptor fd, int offset, int length);
	private static native int nativePlay(int soundId, float volume, float stereo);
	private static native void nativeStop(int soundId);
	private static native void nativeStopAll();
	private static native void nativeRelease();

	static
	{
		System.loadLibrary("openslplayer");
	}

	/**
	 * Initializes the OpenSL sound player.
	 * @param MaxStreams sets the maximum number of simultaneous streams.
	 * @param Stereo indicates if the sounds to play have two channels (stereo/true) or only one (mono/false).
	 */
	public void Initialize(int MaxStreams, boolean Stereo)
	{
		if (Stereo)
			createEngine(MaxStreams, 2);
		else
			createEngine(MaxStreams, 1);
	}

	/**
	 * Loads a sound file located in the Assets folder and returns the SoundID required by the Play function.
	 * The expected format is PCM 16 bits 44.1 Khz.
	 */
	public int LoadSoundFromAssets(BA ba, String FileName) throws IOException
	{
		AssetFileDescriptor afd = ba.context.getAssets().openFd(FileName);
		int res = 0;
		if (afd != null) {
			FileDescriptor fd = afd.getFileDescriptor();
			int off = (int) afd.getStartOffset();
			int len = (int) afd.getLength();
			res = load(fd, off, len);
		}
		return res;
	}

	/**
	 * Loads a sound file located on the device's storage media and returns the SoundID required by the Play function.
	 * The expected format is PCM 16 bits 44.1 Khz.
	 */
	public int LoadSoundFromDisk(BA ba, String FileName) throws IOException
	{
		File file = new File(FileName);
		FileInputStream fileIS = new FileInputStream(file);
		int res = 0;
		if (fileIS != null) {
			FileDescriptor fd = fileIS.getFD();
			int off = 0;
			int len = (int) file.length();
			res = load(fd, off, len);
			fileIS.close();
		}
		return res;
	}

	/**
	 * Plays the sound with the specified SoundID.
	 * Returns True if the sound is playing, False if no stream is available.
	 * @param Volume 0 (mute) to 1 (loud)
	 * @param StereoPosition -1 (left) to 1 (right)
	 */
	public boolean Play(int SoundID, float Volume, float StereoPosition)
	{
		return (nativePlay(SoundID, Math.max(0, Math.min(Volume, 1)), Math.max(-1, Math.min(StereoPosition, 1)) * 1000) > 0);
	}

	/**
	 * Stops the sound with the specified SoundID.
	 */
	public void Stop(int SoundID)
	{
		if (SoundID > 0)
			nativeStop(SoundID);
	}

	/**
	 * Stops all sounds currently playing.
	 */
	public void StopAll()
	{
		nativeStopAll();
	}

	/**
	 * Frees the resources allocated to this player.
	 */
	public void Release()
	{
		nativeRelease();
	}
}
